set linesize 32000
set pagesize 1000
set long 2000000000
set longchunksize 1000
set head off;
set verify off;
set termout off;
 
column u new_value us noprint;
column n new_value ns noprint;
 
select name n from v$database;
select user u from dual;
set sqlprompt &ns:&us>

set head on
set echo on
set termout on
set trimspool on

set serveroutput on size 1000000

spool &ns.tswh.log

set define off

DROP TABLE ESTIMATE_TABLES;

CREATE TABLE ESTIMATE_TABLES
(
  OWNER       VARCHAR2(30)                 NOT NULL,
  TABLE_NAME  VARCHAR2(30)                 NOT NULL,
  PERCENT     NUMBER
);

CREATE INDEX ESTIMATE_TABLES_IDX ON ESTIMATE_TABLES
(OWNER, TABLE_NAME);

DROP TABLE HIST_COLUMNS;

CREATE TABLE HIST_COLUMNS
(
  OWNER        VARCHAR2(30),
  TABLE_NAME   VARCHAR2(30),
  COLUMN_NAME  VARCHAR2(30)
);


CREATE INDEX HIST_COLUMNS_IDX ON HIST_COLUMNS
(OWNER, TABLE_NAME, COLUMN_NAME);

-- Create stattab

declare
  dbname varchar2(2000);
  curschema varchar2(2000);
begin

  select name into dbname from v$database;
  select user into curschema from dual;
  
  execute immediate 'DROP TABLE '||curschema||'.'||'STATTAB_'||dbname;
end;
/

declare
  dbname varchar2(2000);
  curschema varchar2(2000);
begin

  select name into dbname from v$database;
  select user into curschema from dual;
  
  DBMS_STATS.CREATE_STAT_TABLE (curschema,'STATTAB_'||dbname);
end;
/

CREATE OR REPLACE procedure estimate_granularity(
       towner varchar2,
       tname varchar2,
       part varchar2,
       est_percent out NUMBER,
       gran out VARCHAR2) is

   cursor est_tabs_curs(own varchar2,tab varchar2) is
            select
                percent
            from estimate_tables
            where
               owner=own and
               table_name=tab;

   est_tabs_rec est_tabs_curs%rowtype;

 
   table_size number;
   num_rows number;
   
begin

-- Calculate default estimate percent based on table size - otherwise will be overridden
-- by estimate percent on estimate_tables.  

    select count(*) into num_rows from dba_tables t where t.owner=towner and t.table_name=tname;
    if num_rows > 0 then  
      select t.blocks*p.value bytes into table_size from dba_tables t,v$parameter p
      where p.name='db_block_size' and t.owner=towner and
      t.table_name=tname;
	else
	  table_size := 0;
	end if;
	
-- new code to handle table_size = 0.  might be large new table and you don't want to do full 
-- estimate percent.

    if table_size = 0 or table_size is null then  
        select count(*) into num_rows from dba_segments s where s.owner=towner and s.segment_name=tname and
        s.segment_type in ('TABLE','TABLE PARTITION');
        if num_rows > 0 then  
            select sum(s.blocks)*p.value bytes into table_size from dba_segments s,v$parameter p
            where p.name='db_block_size' and s.owner=towner and
            s.segment_name=tname and s.segment_type in ('TABLE','TABLE PARTITION')
            group by p.value;
        end if;
    end if;

-- Set granularity to PARTITION unless is for table as a whole

    gran := 'PARTITION';
    if part is null then
         gran := 'GLOBAL';
    end if;
   
    est_percent := NULL;
 
-- all tables < 1 gig do 100%
-- >= 1 gig do 1/10*gig percent

    if table_size >= 1024*1024*1024 then
        est_percent := (1024*1024*1024)/(10*table_size);
    end if;

-- if table is listed in estimate_tables use that percentage

    open est_tabs_curs(towner,tname);
    loop
        fetch est_tabs_curs into est_tabs_rec;
        exit when est_tabs_curs%notfound;
            est_percent:=est_tabs_rec.percent;
        exit;
    end loop;
    close est_tabs_curs;

end;
/
show errors

-- added no_invalidate option
-- will not work on 8i 
-- Bobby Durrett 11/19/2008

CREATE OR REPLACE procedure tab_stats_with_hists (
       towner varchar2,
       tname varchar2,
       part varchar2) is

    cursor hist_cols_curs(own varchar2,tab varchar2) is
        select
            column_name
        from hist_columns
        where
           owner=own and
           table_name=tab;
    hist_cols_rec hist_cols_curs%rowtype;
    column_options varchar2(32767);
    saved_column_options varchar2(32767);
    num_chars_col_opts number;

    cursor dba_cols_curs(own varchar2,tab varchar2) is
        select
            column_name
        from dba_tab_cols
        where
           owner=own and
           table_name=tab and
           (owner,table_name,column_name) not in
               (select
                owner,table_name,column_name
                from hist_columns
                where
                    owner=own and
                    table_name=tab);
    dba_cols_rec dba_cols_curs%rowtype;

    dbname VARCHAR2(9);

    est_percent NUMBER;
    gran VARCHAR2(30);

   unusable_index EXCEPTION;
   PRAGMA EXCEPTION_INIT(unusable_index, -20000);

   unusable_index2 EXCEPTION;
   PRAGMA EXCEPTION_INIT(unusable_index2, -12801);
   
   resource_busy EXCEPTION;
   PRAGMA EXCEPTION_INIT(resource_busy, -54);
   
   needs_subpartition_stats boolean;
   subpart_count number;
  
begin

-- get estimate percentage and granularity

    estimate_granularity(towner,tname,part,est_percent,gran);
    
-- figure out if we need to regather stats at subpartition level
-- after gathering at the partition level

    needs_subpartition_stats := FALSE;
    
    if gran = 'PARTITION' then
      select SUBPARTITION_COUNT into subpart_count
      from dba_tab_partitions 
      where 
      table_owner=towner and
      table_name=tname and
      partition_name=part;
      if subpart_count > 0 then
        needs_subpartition_stats := TRUE;
      end if;
    end if;
    
    select name into dbname from v$database;

    column_options := 'FOR COLUMNS ';
-- Number of characters of start of options string
    num_chars_col_opts := 12;

    open hist_cols_curs(towner,tname);
    loop
        fetch hist_cols_curs into hist_cols_rec;
        exit when hist_cols_curs%notfound;

-- add in size of column name plus size of ' SIZE 250,' which is added to column name
        num_chars_col_opts := num_chars_col_opts + length(hist_cols_rec.column_name) + 10;

-- check to see if size of options string will be too big.  if so, exit loop so we can skip
-- histograms on this one.
        if num_chars_col_opts > 32767 then
        	num_chars_col_opts := -1;
            exit;
        end if;

-- Add columns to options string
        column_options := column_options || hist_cols_rec.column_name || ' SIZE 250,';
    end loop;
    close hist_cols_curs;

-- 12 or less means don't do histograms on columns - either not in table or options string
-- is too long
    if num_chars_col_opts <= 12 then
        DBMS_STATS.GATHER_TABLE_STATS (towner,
                                       tname,
                                       part,
                                       cascade=>TRUE,
				       degree=>5,
				       granularity=>gran,
                                       estimate_percent=>est_percent,
                                       method_opt=>'FOR ALL COLUMNS SIZE 1',
                                       stattab=>'STATTAB_'||dbname,
                                       statid=>substr(towner,1,14) || TO_CHAR(SYSDATE,'yymmddhh24miss'),
                                       no_invalidate=>FALSE);

        DBMS_OUTPUT.put_line(tname||' '||part);
        if needs_subpartition_stats = TRUE then
          DBMS_STATS.GATHER_TABLE_STATS (towner,
                                         tname,
                                         part,
                                         cascade=>TRUE,
	      			         degree=>5,
				         granularity=>'SUBPARTITION',
                                         estimate_percent=>est_percent,
                                         method_opt=>'FOR ALL COLUMNS SIZE 1',
                                         stattab=>'STATTAB_'||dbname,
                                         statid=>substr(towner,1,14) || TO_CHAR(SYSDATE,'yymmddhh24miss'),
                                         no_invalidate=>FALSE);
         
        end if;
        return;
    end if;


    saved_column_options := column_options;

    open dba_cols_curs(towner,tname);
    loop
        fetch dba_cols_curs into dba_cols_rec;
        exit when dba_cols_curs%notfound;

 -- ' SIZE 1,' is 8 characters.
        num_chars_col_opts := num_chars_col_opts + length(dba_cols_rec.column_name) + 8;

        if num_chars_col_opts > 32767 then
        	num_chars_col_opts := -1;
            exit;
        end if;

        column_options := column_options || dba_cols_rec.column_name || ' SIZE 1,';
    end loop;
    close dba_cols_curs;

-- -1 means options string is too long with all the non-histogram columns added.
-- In this case gather stats in two steps.  First for all columns, then for the
-- histogram columns.
    if num_chars_col_opts <= 12 then
        DBMS_STATS.GATHER_TABLE_STATS (towner,
                                       tname,
                                       part,
                                       cascade=>TRUE,
				       degree=>5,
				       granularity=>gran,
                                       estimate_percent=>est_percent,
                                       method_opt=>'FOR ALL COLUMNS SIZE 1',
                                       stattab=>'STATTAB_'||dbname,
                                       statid=>substr(towner,1,14) || TO_CHAR(SYSDATE,'yymmddhh24miss'),
                                       no_invalidate=>FALSE);

        DBMS_OUTPUT.put_line(tname||' '||part);
        if needs_subpartition_stats = TRUE then
          DBMS_STATS.GATHER_TABLE_STATS (towner,
                                         tname,
                                         part,
                                         cascade=>TRUE,
				         degree=>5,
				         granularity=>'SUBPARTITION',
                                         estimate_percent=>est_percent,
                                         method_opt=>'FOR ALL COLUMNS SIZE 1',
                                         stattab=>'STATTAB_'||dbname,
                                         statid=>substr(towner,1,14) || TO_CHAR(SYSDATE,'yymmddhh24miss'),
                                         no_invalidate=>FALSE);
        end if;


-- Restore options with just columns that need histograms

        column_options := saved_column_options;
    end if;

    column_options := substr(column_options,1,length(column_options)-1); -- Trim off last comma

-- Do histograms on columns

    DBMS_STATS.GATHER_TABLE_STATS (towner,
                                   tname,
                                   part,
                                   cascade=>TRUE,
 				   degree=>5,
                                   granularity=>gran,
                                   estimate_percent=>est_percent,
                                   method_opt=>column_options,
                                   stattab=>'STATTAB_'||dbname,
                                   statid=>substr(towner,1,14) || TO_CHAR(SYSDATE,'yymmddhh24miss'),
                                   no_invalidate=>FALSE);

    DBMS_OUTPUT.put_line(tname||' '||part);
    
    if needs_subpartition_stats = TRUE then
      DBMS_STATS.GATHER_TABLE_STATS (towner,
                                     tname,
                                     part,
                                     cascade=>TRUE,
 				     degree=>5,
                                     granularity=>'SUBPARTITION',
                                     estimate_percent=>est_percent,
                                     method_opt=>column_options,
                                     stattab=>'STATTAB_'||dbname,
                                     statid=>substr(towner,1,14) || TO_CHAR(SYSDATE,'yymmddhh24miss'),
                                     no_invalidate=>FALSE);
   
    end if;
        
    while length(column_options) > 80 loop
        DBMS_OUTPUT.put_line(substr(column_options,1,80));
        column_options:=substr(column_options,81,length(column_options)-80);
    end loop;
    if length(column_options) > 0 then
        DBMS_OUTPUT.put_line(column_options);
    end if;
    EXCEPTION
    WHEN unusable_index THEN
        DBMS_OUTPUT.put_line('Unusable index on '||towner|| ' ' ||tname|| ' ' ||part);
    WHEN unusable_index2 THEN
        DBMS_OUTPUT.put_line('Unusable index on '||towner|| ' ' ||tname|| ' ' ||part);
    WHEN resource_busy THEN
        DBMS_OUTPUT.put_line('Resource busy on '||towner|| ' ' ||tname|| ' ' ||part);
end;
/
show errors

spool off

